@@ -1,11 +1,11 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2004 Nokia. All rights reserved.
+ * Copyright (C) 2004-2006 Nokia Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * published by the Free Software Foundation; version 2 of the
+ * License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -19,23 +19,32 @@
  */
 
 #include <config.h>
-
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
 #include <libgnomevfs/gnome-vfs-method.h>
+#include <libgnomevfs/gnome-vfs-module.h>
+#include <libgnomevfs/gnome-vfs-module-shared.h>
 #include <gw-obex.h>
-#include <pthread.h>
-
 #include <obex-vfs-utils/ovu-cap-server.h>
+#include <obex-vfs-utils/ovu-cap-parser.h>
 
 #include "om-fl-parser.h"
 #include "om-utils.h"
 #include "om-dbus.h"
 #include "om-vfs-utils.h"
 
+#define DBUS_API_SUBJECT_TO_CHANGE
+#include <dbus/dbus.h>
+
+/* For debug output (like errors/warnings). */
+#define d(x) x
+
+/* For really verbose debut output (like for every read/write). */
+#define dv(x) 
+
 /* Lock for all manipulations of the connection hash table. This includes
  * ref/unref of ObexConnection objects.
  */
@@ -55,7 +64,8 @@ typedef struct {
 	/* The device we're connected to if we're connected. */
 	gchar  *connected_dev;
 
-	GwObex *obex;
+	GMainContext *context;
+	GwObex       *obex;
 
 	gchar  *current_dir;
 	GList  *current_listing;
@@ -69,23 +79,18 @@ typedef struct {
 	guint   ref_count;
 
 	gboolean disconnected;
+
+	/* The current file name */
+	gchar   *file_name;
 } ObexConnection;
 
 typedef struct {
 	GnomeVFSURI      *uri;
 
 	GnomeVFSOpenMode  mode;
+	gboolean          has_eof;
 
-	/* Indicates that the file needs to be pushed to the remote device when
-	 * we close it.
-	 */
-	gboolean          written;
-
-	/* We do all reading/writing from/to a temporary file on the local
-	 * side.
-	 */
-	gint              temp_fd;
-	gchar            *temp_filename;
+	GwObexXfer       *xfer;
 } FileHandle;
 
 typedef struct {
@@ -106,6 +111,12 @@ typedef struct {
 	GList *handles; /* list of MonitorHandle* */
 } MonitorList;
 
+/* The time an idle obex connection is kept alive for connction reuse. */
+#define OBEX_TIMEOUT 20000
+
+
+static void       om_connection_set_current_uri  (ObexConnection    *conn,
+						  const GnomeVFSURI *uri);
 
 static void             om_connection_unref      (ObexConnection    *conn);
 static void             om_connection_free       (ObexConnection    *conn);
@@ -118,10 +129,10 @@ static void             om_disconnect_cb         (GwObex             *ctx,
 static gboolean         om_cancel_cb             (GnomeVFSContext    *context);
 static void             om_set_cancel_context    (ObexConnection     *conn,
 						  GnomeVFSContext    *context);
-static GnomeVFSResult   om_close_file_helper     (gint                fd,
-						  GnomeVFSContext    *context);
 static void             om_directory_handle_free (DirectoryHandle    *handle);
-static FileHandle *     om_file_handle_new       (const GnomeVFSURI  *uri);
+static FileHandle *     om_file_handle_new       (const GnomeVFSURI  *uri,
+						  GnomeVFSOpenMode    mode,
+						  GwObexXfer         *xfer);
 static void             om_file_handle_free      (FileHandle         *handle);
 static GnomeVFSResult   om_get_folder_listing    (ObexConnection     *conn,
 						  const GnomeVFSURI  *uri,
@@ -131,16 +142,6 @@ static GnomeVFSResult   om_get_folder_listing    (ObexConnection     *conn,
 static GnomeVFSResult   om_chdir_to_uri          (ObexConnection     *conn, 
 						  const GnomeVFSURI  *uri,
 						  gboolean            to_parent);
-static GnomeVFSResult   om_put_file              (ObexConnection     *conn,
-						  const gchar        *local_filename,
-						  const GnomeVFSURI  *remote_uri);
-static void             remove_write_file_handle (FileHandle         *handle, 
-						  GnomeVFSContext    *context);
-GnomeVFSMethod *        vfs_module_init          (const char         *method_name,
-						  const char         *args);
-void                    vfs_module_shutdown      (GnomeVFSMethod     *method);
-
-
 static void             om_notify_monitor        (GnomeVFSURI        *uri,
 						  GnomeVFSMonitorEventType event);
 
@@ -151,6 +152,45 @@ static GnomeVFSResult   do_set_file_info         (GnomeVFSMethod          *metho
 						  GnomeVFSSetFileInfoMask  mask,
 						  GnomeVFSContext         *context);
 
+#if 0
+/* Leave this is since it's useful for debugging. */
+static const gchar *
+event_to_string (GnomeVFSMonitorEventType event)
+{
+	switch (event) {
+	case GNOME_VFS_MONITOR_EVENT_CHANGED:
+		return "changed";
+		
+	case GNOME_VFS_MONITOR_EVENT_DELETED:
+		return "deleted";
+		
+	case GNOME_VFS_MONITOR_EVENT_CREATED:
+		return "created";
+
+	case GNOME_VFS_MONITOR_EVENT_STARTEXECUTING:
+	case GNOME_VFS_MONITOR_EVENT_STOPEXECUTING:
+	case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
+		break;
+	}
+	
+	return "<not supported>";
+}
+#endif
+
+static void 
+om_connection_set_current_uri (ObexConnection    *conn,
+			       const GnomeVFSURI *uri)
+{
+	g_free (conn->file_name);
+	
+	if (uri) {
+		conn->file_name = gnome_vfs_uri_to_string (uri, 
+							   GNOME_VFS_URI_HIDE_NONE);
+	} else {
+		conn->file_name = NULL;
+	}
+}
+							      
 static void
 om_connection_unref (ObexConnection *conn)
 {
@@ -184,7 +224,7 @@ om_connection_unref (ObexConnection *conn)
 	 * to use the same connection.
 	 */
 	if (conn->timeout_id == 0) {
-		conn->timeout_id = g_timeout_add (10000, 
+		conn->timeout_id = g_timeout_add (OBEX_TIMEOUT, 
 						  (GSourceFunc) om_connection_timed_out, 
 						  conn);
 	}
@@ -197,16 +237,22 @@ om_connection_free (ObexConnection *conn)
 {
 	if (conn->obex) {
 		gw_obex_close (conn->obex);
+	}
+	
+	if (conn->connected_dev) {
+		d(g_printerr ("obex: om_connection_free calls disconnect\n"));
+		om_dbus_disconnect_dev (conn->connected_dev);
+		g_free (conn->connected_dev);
+		conn->connected_dev = NULL;
+	}
 
-		if (conn->connected_dev) {
-			om_dbus_disconnect_dev (conn->connected_dev);
-			g_free (conn->connected_dev);
-			conn->connected_dev = NULL;
-		}
+	if (conn->context) {
+		g_main_context_unref (conn->context);
 	}
 	
 	g_free (conn->dev);
 	g_free (conn->current_dir);
+	g_free (conn->file_name);
 
 	if (conn->current_listing) {
 		gnome_vfs_file_info_list_free (conn->current_listing);
@@ -219,6 +265,7 @@ om_connection_free (ObexConnection *conn)
 static gboolean 
 om_connection_timed_out (ObexConnection *conn)
 {
+	d(g_printerr ("Connection timed out\n"));
 	g_mutex_lock (conn_hash_mutex);
 
 	if (conn->ref_count > 0) {
@@ -260,18 +307,19 @@ om_connection_reset (ObexConnection *conn)
 	gint            error;
 
 	if (conn->obex) {
-		gw_obex_close (conn->obex);
+		gw_obex_close (conn->obex); 
 
-		if (conn->connected_dev) {
-			om_dbus_disconnect_dev (conn->connected_dev);
-			g_free (conn->connected_dev);
-			conn->connected_dev = NULL;
-		}
-		
 		conn->obex = NULL;
 		conn->disconnected = TRUE;
 	}
 
+	if (conn->connected_dev) {
+		d(g_printerr ("obex: om_connection_reset calls disconnect\n"));
+		om_dbus_disconnect_dev (conn->connected_dev);
+		g_free (conn->connected_dev);
+		conn->connected_dev = NULL;
+	}
+	
 	if (strncmp (conn->dev, "/dev/rfcomm", 11) == 0) {
 		real_dev = g_strdup (conn->dev);
 	} else {
@@ -282,23 +330,28 @@ om_connection_reset (ObexConnection *conn)
 	}
 
 	/* New connection, set it up. */
+	if (!conn->context) {
+		conn->context = g_main_context_new ();
+	}
+
 	error = 0;
 	obex = gw_obex_setup_dev (real_dev, OBEX_FTP_UUID, 
-                                  OBEX_FTP_UUID_LEN, NULL, &error);
+                                  OBEX_FTP_UUID_LEN, conn->context, &error);
 	if (obex == NULL) {
+		d(g_printerr ("obex: om_connection_reset calls disconnect (error)\n"));
+		om_dbus_disconnect_dev (real_dev);
 		g_free (real_dev);
-
+		
 		return om_utils_obex_error_to_vfs_result (error);
 	}
 	
 	conn->connected_dev = real_dev;
 	conn->obex = obex;
 	conn->disconnected = FALSE;
-	
+
 	gw_obex_set_disconnect_callback (conn->obex,
 					 (gw_obex_disconnect_cb_t) om_disconnect_cb,
 					 conn);
-
 	g_free (conn->current_dir);
 	conn->current_dir = NULL;
 
@@ -384,6 +437,8 @@ om_get_connection (const GnomeVFSURI *uri, GnomeVFSResult *result)
 static void
 om_disconnect_cb (GwObex *ctx, ObexConnection *conn)
 {
+	d(g_printerr ("obex: Disconnect callback called\n"));
+
 	conn->disconnected = TRUE;
 }
 
@@ -396,6 +451,10 @@ om_cancel_cb (GnomeVFSContext *context)
 	cancellation = gnome_vfs_context_get_cancellation (context);
 
 	cancel = gnome_vfs_cancellation_check (cancellation);
+
+	if (cancel) {
+		d(g_printerr ("obex: Cancellation check => CANCEL\n"));
+	}
 	
 	return cancel;
 }
@@ -425,22 +484,14 @@ om_set_cancel_context (ObexConnection  *conn,
 	}
 }
 
-static GnomeVFSResult
-om_close_file_helper (gint             fd,
-		      GnomeVFSContext *context)
+static void
+om_connection_check_gwobex_error (ObexConnection *conn,
+				  gint            error)
 {
-	gint retval;
-
-	do {
-		retval = close (fd);
-	} while (retval != 0 && errno == EINTR
-		 && !gnome_vfs_context_check_cancellation (context));
-	
-	if (retval != 0) {
-		return gnome_vfs_result_from_errno ();
+	if (error == GW_OBEX_ERROR_TIMEOUT ||
+	    error == GW_OBEX_ERROR_DISCONNECT) {
+		conn->disconnected = TRUE;
 	}
-
-	return GNOME_VFS_OK;
 }
 
 static void
@@ -452,15 +503,17 @@ om_directory_handle_free (DirectoryHandle *handle)
 }
 
 static FileHandle *
-om_file_handle_new (const GnomeVFSURI *uri)
+om_file_handle_new (const GnomeVFSURI *uri,
+		    GnomeVFSOpenMode   mode,
+		    GwObexXfer        *xfer)
 {
 	FileHandle *handle;
 
 	handle = g_new0 (FileHandle, 1);
 
-	handle->uri = gnome_vfs_uri_dup (uri);
-	handle->temp_fd = -1;
-	handle->written = FALSE;
+	handle->uri  = gnome_vfs_uri_dup (uri);
+	handle->mode = mode;
+	handle->xfer = xfer;
 
 	return handle;
 }
@@ -470,6 +523,11 @@ om_file_handle_free (FileHandle *handle)
 {
 	gnome_vfs_uri_unref (handle->uri);
 
+	if (handle->xfer) {
+		gw_obex_xfer_close (handle->xfer, NULL);
+		gw_obex_xfer_free (handle->xfer);
+	}
+
 	g_free (handle);
 }
 
@@ -498,7 +556,10 @@ om_get_folder_listing (ObexConnection     *conn,
 	if (path && conn->current_dir && 
 	    strcmp (path, conn->current_dir) == 0 &&
 	    conn->current_listing != NULL) {
-		*list = gnome_vfs_file_info_list_copy (conn->current_listing);
+		if (list) {
+			*list = gnome_vfs_file_info_list_copy (conn->current_listing);
+		}
+
 		g_free (path);
 		return GNOME_VFS_OK;
 	}
@@ -515,6 +576,9 @@ om_get_folder_listing (ObexConnection     *conn,
 	om_set_cancel_context (conn, context);
 	if (!gw_obex_read_dir (conn->obex, "", &buf, &len, &error)) {
 		om_set_cancel_context (conn, NULL);
+
+		om_connection_check_gwobex_error (conn, error);
+
 		return om_utils_obex_error_to_vfs_result (error);
 	}
 	om_set_cancel_context (conn, NULL);
@@ -522,7 +586,7 @@ om_get_folder_listing (ObexConnection     *conn,
 	retval = om_fl_parser_parse (buf, len, &elements, &gerror);
 	g_free (buf);
 	if (gerror) {  
-		/* g_print ("%s\n", gerror->message); */
+		d(g_printerr ("obex: folder listing parse failed: %s\n", gerror->message));
 		g_error_free (gerror);  
 	}
 
@@ -535,7 +599,9 @@ om_get_folder_listing (ObexConnection     *conn,
 
 	conn->current_listing = gnome_vfs_file_info_list_copy (elements);
 		
-	*list = elements;
+	if (list) {
+		*list = elements;
+	}
 
 	return GNOME_VFS_OK;
 }
@@ -571,6 +637,8 @@ om_chdir_to_uri (ObexConnection    *conn,
 		gint error;
 
 		if (!gw_obex_chdir (conn->obex, l->data, &error)) {
+			om_connection_check_gwobex_error (conn, error);
+			
 			result = om_utils_obex_error_to_vfs_result (error);
 			break;
 		}
@@ -594,94 +662,13 @@ om_chdir_to_uri (ObexConnection    *conn,
 }
 
 static GnomeVFSResult
-om_put_file (ObexConnection    *conn,
-	     const gchar       *local_filename,
-	     const GnomeVFSURI *remote_uri)
-{
-	GnomeVFSResult  result;
-	gchar          *remote_name;
-	gint            error;
-	
-	if (conn->obex == NULL) {
-		return GNOME_VFS_ERROR_CANCELLED;
-	}
-
-	result = om_chdir_to_uri (conn, remote_uri, TRUE);
-	if (result != GNOME_VFS_OK) {
-		return result;
-	}
-		
-	remote_name = gnome_vfs_uri_extract_short_name (remote_uri);
-
-	if (!gw_obex_put_file (conn->obex,
-			       local_filename,
-			       remote_name,
-			       &error)) {
-		g_free (remote_name);
-	
-		return om_utils_obex_error_to_vfs_result (error);
-	}
-
-	om_connection_invalidate_cache (conn);
-	
-	g_free (remote_name);
-
-	return GNOME_VFS_OK;
-}
-
-static GnomeVFSResult
-create_file (ObexConnection        *conn,
-	     GnomeVFSMethodHandle **method_handle,
-	     GnomeVFSURI           *uri,
-	     GnomeVFSOpenMode       mode,
-	     gboolean               exclusive,
-	     guint                  perm,
-	     GnomeVFSContext       *context)
-{
-	gchar          *path;
-	gchar          *filename;
-	gint            fd;
-	FileHandle     *handle;
-	GnomeVFSResult  result;
-
-	/* Note that this function needs a conn that is setup with cancellation
-	 * context etc from the caller.
-	 */
-	
-	/* Exclusive is not supported, since that would mean to we must check if
-	 * the remote side has the file and error out if so.
-	 */
-
-	path = om_utils_get_path_from_uri (uri);
-	if (path == NULL) {
-		return GNOME_VFS_ERROR_INVALID_URI;
-	}
-
-	/* Try to change into the directory to see that it exists */
-	result = om_chdir_to_uri (conn, uri, TRUE);
-	if (result != GNOME_VFS_OK) {
-		return result;
-	}
-	
-	fd = g_file_open_tmp ("om-tmpfile-write-XXXXXX",
-			      &filename,
-			      NULL);  
-
-	if (fd == -1) {
-		g_free (path);
-		return GNOME_VFS_ERROR_INTERNAL;
-	}
-
-	handle = om_file_handle_new (uri);
-
-	handle->temp_fd = fd;
-	handle->temp_filename = filename;
-	handle->mode = mode;
-
-	*method_handle = (GnomeVFSMethodHandle *) handle;
-
-	return GNOME_VFS_OK;
-}
+do_create (GnomeVFSMethod        *method,
+	   GnomeVFSMethodHandle **method_handle,
+	   GnomeVFSURI           *uri,
+	   GnomeVFSOpenMode       mode,
+	   gboolean               exclusive,
+	   guint                  perm,
+	   GnomeVFSContext       *context);
 
 static GnomeVFSResult 
 do_open (GnomeVFSMethod        *method,
@@ -693,15 +680,29 @@ do_open (GnomeVFSMethod        *method,
 	ObexConnection *conn;
 	gint            error;
 	FileHandle     *handle;
-	gint            fd = -1;
-	gchar          *local_filename;
 	gchar          *name;
 	GnomeVFSResult  result;
+	GwObexXfer     *xfer;
+
+	/* Only support either read or write */
+	if ((mode & GNOME_VFS_OPEN_READ) && (mode & GNOME_VFS_OPEN_WRITE)) {
+		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
+	}
 
 	if (!(mode & GNOME_VFS_OPEN_READ) && !(mode & GNOME_VFS_OPEN_WRITE)) {
 		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
 	}
 
+	if (mode & GNOME_VFS_OPEN_WRITE) {
+		return do_create (method,
+				  method_handle, 
+				  uri, 
+				  mode, 
+				  FALSE, /* Ignored */
+				  0,     /* Ignored */
+				  context);
+	}
+
 	conn = om_get_connection (uri, &result);
 	if (conn == NULL) {
 		return result;
@@ -715,80 +716,40 @@ do_open (GnomeVFSMethod        *method,
 		return result;
 	}
 
-	fd = g_file_open_tmp ("om-tmpfile-XXXXXX",
-			      &local_filename,
-			      NULL);  
+	/* Cancellation context is not used beyond this point */
+	om_set_cancel_context (conn, NULL);
 
-	if (fd == -1) {
-		g_free (local_filename);
-		
-		om_set_cancel_context (conn, NULL);
-		om_connection_unref (conn);
-		return GNOME_VFS_ERROR_INTERNAL;
-	}
-
-	*method_handle = NULL;
-	
 	name = gnome_vfs_uri_extract_short_name (uri);
-	if (!gw_obex_get_fd (conn->obex,
-			     fd,
-			     name,
-			     &error)) {
+
+	dv(g_printerr ("do_open: %s\n", name));
+	
+	om_connection_set_current_uri (conn, uri);
+
+	xfer = gw_obex_get_async (conn->obex, name, NULL, &error);
+	if (!xfer) {
+		om_connection_set_current_uri (conn, NULL);
+
+		om_connection_check_gwobex_error (conn, error);
 
 		result = om_utils_obex_error_to_vfs_result (error);
 
-		/* Note: We can't check for NOT_FOUND here since various Obex
-		 * implementations return a different error here, so just check
-		 * if there is an error.
-		 */
-		if ((mode & GNOME_VFS_OPEN_WRITE) && result != GNOME_VFS_OK) {
-			/* Create the file if write mode and the file doesn't
-			 * exist.
-			 */
-			result = create_file (conn,
-					      method_handle,
-					      uri,
-					      mode,
-					      FALSE, /* exclusive */
-					      0,     /* perms */
-					      context);
-		}
+		/* Clean up local variables */
+		g_free (name);
+		om_connection_unref (conn);
 
-		if (result != GNOME_VFS_OK) {
-			om_close_file_helper (fd, context);
-			unlink (local_filename);
-			
-			g_free (local_filename);
-			g_free (name);
-			
-			om_set_cancel_context (conn, NULL);
-			
-			om_connection_unref (conn);
-			return result;
-		}
+		return result;
 	}
 
-	om_set_cancel_context (conn, NULL);
+	/* Always use the Xfer object in blocking mode */
+	gw_obex_xfer_set_blocking (xfer, TRUE);
 
 	g_free (name);
 
 	om_connection_unref (conn);
 
-	/* Rewind the file so we can read it again. */
-	if (lseek (fd, 0, SEEK_SET) == -1) {
-		g_free (local_filename);
-		return gnome_vfs_result_from_errno ();
-	}
+	handle = om_file_handle_new (uri, mode, xfer);
 
-	if (*method_handle == NULL) {
-		handle = om_file_handle_new (uri);
-		handle->mode = mode;
-		
-		handle->temp_fd = fd;
-		handle->temp_filename = local_filename;
-	
-		*method_handle = (GnomeVFSMethodHandle *) handle;
-	}
+	*method_handle = (GnomeVFSMethodHandle *) handle;
 	
 	return GNOME_VFS_OK;
 }
@@ -804,7 +765,11 @@ do_create (GnomeVFSMethod        *method,
 {
 	ObexConnection *conn;
 	GnomeVFSResult  result;
-	
+	FileHandle     *handle;
+	gchar          *name;
+	GwObexXfer     *xfer;
+	gint            error;
+
 	if (!(mode & GNOME_VFS_OPEN_WRITE)) {
 		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
 	}
@@ -821,29 +786,48 @@ do_create (GnomeVFSMethod        *method,
 		om_connection_unref (conn);
 		return result;
 	} 
-	
-	result = create_file (conn, method_handle, uri,
-			      mode, exclusive, perm, context);
 
+	/* Cancellation context is not used beyond this point */
 	om_set_cancel_context (conn, NULL);
+
+	name = gnome_vfs_uri_extract_short_name (uri);
+	
+	xfer = gw_obex_put_async (conn->obex, name, NULL, 
+				  GW_OBEX_UNKNOWN_LENGTH, /* Object size */
+				  -1, /* mtime */
+				  &error);
+
+	if (!xfer) {
+		om_connection_set_current_uri (conn, NULL);
+
+		om_connection_check_gwobex_error (conn, error);
+
+		result = om_utils_obex_error_to_vfs_result (error);
+		
+		/* Clean up local variables */
+		g_free (name);
+		om_connection_unref (conn);
+
+		return result;
+	}
+
+	/* Always use the Xfer object in blocking mode */
+	gw_obex_xfer_set_blocking (xfer, TRUE);
+
+	g_free (name);
+
 	om_connection_unref (conn);
 
-	return result;
-}
+	handle = om_file_handle_new (uri, mode, xfer);
 
-static void
-remove_write_file_handle (FileHandle *handle, GnomeVFSContext *context)
-{
-	if (handle->temp_fd != -1) {
-		/* We don't bail out if this can't be closed. */
-		om_close_file_helper (handle->temp_fd, context);
-	}
-	
-	if (handle->temp_filename) {
-		unlink (handle->temp_filename);
-	}
-	
-	om_file_handle_free (handle);
+	*method_handle = (GnomeVFSMethodHandle *) handle;
+
+	/* We don't emit this here, since the file doesn't exist on the phone
+	 * until it's been closed.
+	 */
+	/*om_notify_monitor (uri, GNOME_VFS_MONITOR_EVENT_CREATED);*/
+
+	return GNOME_VFS_OK;
 }
 
 static GnomeVFSResult
@@ -853,37 +837,55 @@ do_close (GnomeVFSMethod       *method,
 {
 	FileHandle     *handle;
 	GnomeVFSResult  result;
+	ObexConnection *conn;
+	gint            error;
+	gboolean        success;
+	GnomeVFSURI    *notify_uri;
 
 	result = GNOME_VFS_OK;
 	handle = (FileHandle *) method_handle;
 
-	if (handle->written) {
-		ObexConnection *conn;
-		
-		g_assert (handle->temp_filename != NULL);
-		g_assert (handle->uri != NULL);
-		
-		conn = om_get_connection (handle->uri, &result);
-		if (conn == NULL) {
-			remove_write_file_handle (handle, context);
-			return result;
-		}
-	
-		om_set_cancel_context (conn, context);
-		result = om_put_file (conn,
-				      handle->temp_filename,
- 				      handle->uri);
-		om_set_cancel_context (conn, NULL);
+	/* Cancellation context is not used in this function */
 
-		if (result == GNOME_VFS_OK) {
-			om_notify_monitor (handle->uri, GNOME_VFS_MONITOR_EVENT_CREATED);
-		}
-		
-		om_connection_unref (conn);
+	conn = om_get_connection (handle->uri, &result);
+	if (conn == NULL) {
+		return result;
 	}
 
-	remove_write_file_handle (handle, context);
-	return result;
+	dv(g_printerr ("do_close: calling gw_obex_xfer_close()\n"));
+
+	if (handle->xfer) {
+		success = gw_obex_xfer_close (handle->xfer, &error);
+		gw_obex_xfer_free (handle->xfer);
+
+		handle->xfer = NULL;
+	} else {
+		success = TRUE;
+	}
+
+	if (success && handle->mode & GNOME_VFS_OPEN_WRITE) {
+		notify_uri = gnome_vfs_uri_ref (handle->uri);
+	} else {
+		notify_uri = NULL;
+	}
+	
+	om_file_handle_free (handle);
+
+	om_connection_unref (conn);
+
+	if (!success) {
+		return om_utils_obex_error_to_vfs_result (error);
+	}
+
+	/* The file does not exist in the phone's folder listing until it's
+	 * fully created and closed, that is why we emit this here.
+	 */
+	if (notify_uri) {
+		om_notify_monitor (notify_uri, GNOME_VFS_MONITOR_EVENT_CREATED);
+		gnome_vfs_uri_unref (notify_uri);
+	}
+	
+	return GNOME_VFS_OK;
 }
 
 static GnomeVFSResult
@@ -894,82 +896,65 @@ do_read (GnomeVFSMethod       *method,
 	 GnomeVFSFileSize     *bytes_read,
 	 GnomeVFSContext      *context)
 {
-	FileHandle *handle;
-	gint        read_val;
+	FileHandle     *handle;
+	gint            error;
+	gint            b_read;
+	ObexConnection *conn;
+	GnomeVFSResult  result;
 
 	handle = (FileHandle *) method_handle;
-	
-	do {
-		read_val = read (handle->temp_fd, buffer, num_bytes);
-	} while (read_val == -1 && errno == EINTR
-	         && !gnome_vfs_context_check_cancellation (context));
 
-	if (read_val == -1) {
-		*bytes_read = 0;
-		return gnome_vfs_result_from_errno ();
-	} else {
-		*bytes_read = read_val;
+	/* Cancellation context is not used in this function */
 
-		/* Getting 0 from read() means EOF. */
-		if (read_val == 0) {
+	if (!handle->xfer) {
+		if (handle->has_eof) {
 			return GNOME_VFS_ERROR_EOF;
+		} else {
+			return GNOME_VFS_ERROR_NOT_OPEN;
 		}
 	}
 
-	return GNOME_VFS_OK;
-}
-
-static GnomeVFSResult
-do_seek (GnomeVFSMethod       *method,
-	 GnomeVFSMethodHandle *method_handle,
-	 GnomeVFSSeekPosition  whence,
-	 GnomeVFSFileOffset    offset,
-	 GnomeVFSContext      *context)
-{
-	FileHandle *handle;
-	gint        unix_whence;
-
-	handle = (FileHandle *) method_handle;
-
-	switch (whence) {
-	case GNOME_VFS_SEEK_START:
-		unix_whence = SEEK_SET;
-		break;
-	case GNOME_VFS_SEEK_CURRENT:
-		unix_whence = SEEK_CUR;
-		break;
-	case GNOME_VFS_SEEK_END:
-		unix_whence = SEEK_END;
-		break;
-	default:
-		g_warning ("Invalid seek position type.");
-		unix_whence = SEEK_SET;
-		break;
+	conn = om_get_connection (handle->uri, &result);
+	if (conn == NULL) {
+		return result;
 	}
 
-	if (lseek (handle->temp_fd, offset, unix_whence) == -1) {
-		return gnome_vfs_result_from_errno ();
+        dv(g_printerr ("Trying to read: %d bytes\n", (int) num_bytes));
+
+	b_read = 0;
+	if (!gw_obex_xfer_read (handle->xfer, (char *) buffer, num_bytes,
+				&b_read, &error)) {
+		d(g_printerr ("Read failed: %d bytes read (gwobex error: %d)\n",
+			      b_read, error));
+
+		om_connection_check_gwobex_error (conn, error);
+
+		om_connection_unref (conn);
+
+		return om_utils_obex_error_to_vfs_result (error);
 	}
+
+	dv(g_printerr ("%d bytes read\n", b_read));
+	*bytes_read = b_read;
+
+	om_connection_unref (conn);
+
+	if (b_read == 0) {
+		/* We close the obex handle when the file is read
+		 * completely. The reason for this is that you can then leave a
+		 * file handle open without blocking other obex operations. If
+		 * you try to read again after this, you will just get an EOF
+		 * back.
+		 */
+		gw_obex_xfer_close (handle->xfer, &error);
+		gw_obex_xfer_free (handle->xfer);
+		
+		handle->xfer = NULL;
+		handle->has_eof = TRUE;
 	
-	return GNOME_VFS_OK;
-}
-
-static GnomeVFSResult
-do_tell (GnomeVFSMethod       *method,
-	 GnomeVFSMethodHandle *method_handle,
-	 GnomeVFSFileOffset   *offset_return)
-{
-	FileHandle *handle;
-	off_t       offset;
-
-	handle = (FileHandle *) method_handle;
-
-	offset = lseek (handle->temp_fd, 0, SEEK_CUR);
-	if (offset == -1) {
-		return gnome_vfs_result_from_errno ();
+		return GNOME_VFS_ERROR_EOF;
 	}
 
-	*offset_return = offset;
 	return GNOME_VFS_OK;
 }
 
@@ -981,29 +966,56 @@ do_write (GnomeVFSMethod       *method,
 	  GnomeVFSFileSize     *bytes_written,
 	  GnomeVFSContext      *context)
 {
-	FileHandle *handle;
-	gint        write_val;
+	FileHandle     *handle;
+	gint            error;
+	gint            b_written;
+	ObexConnection *conn;
+	GnomeVFSResult  result;
 
 	handle = (FileHandle *) method_handle;
 
+	/* Cancellation context is not used here */
+
+	if (!handle->xfer) {
+		return GNOME_VFS_ERROR_NOT_OPEN;
+	}
+
 	if (!(handle->mode & GNOME_VFS_OPEN_WRITE)) {
 		return GNOME_VFS_ERROR_READ_ONLY;
 	}
-	
-	do {
-		write_val = write (handle->temp_fd, buffer, num_bytes);
-	} while (write_val == -1 && errno == EINTR
-		 && !gnome_vfs_context_check_cancellation (context));
-	
-	if (write_val == -1) {
-		*bytes_written = 0;
-		return gnome_vfs_result_from_errno ();
-	} else {
-		*bytes_written = write_val;
-		handle->written = TRUE;
-		
-		return GNOME_VFS_OK;
+
+	conn = om_get_connection (handle->uri, &result);
+	if (conn == NULL) {
+		return result;
 	}
+
+        dv(g_printerr ("Trying to write: %d bytes\n", (int) num_bytes));
+
+	b_written = 0;
+	if (!gw_obex_xfer_write (handle->xfer, buffer, num_bytes, 
+				 &b_written, &error)) {
+		d(g_printerr ("Write failed: %d bytes written (gwobex error: %d)\n",
+			      b_written, error));
+
+		om_connection_check_gwobex_error (conn, error);
+
+		om_connection_unref (conn);
+
+		return om_utils_obex_error_to_vfs_result (error);
+	}
+	
+	dv(g_printerr ("%d bytes written\n", b_written));
+	*bytes_written = b_written;
+
+	om_connection_unref (conn);
+
+	/* We don't emit this since the file doesn't exist on the phone until
+	 * it's closed. Having events happen on non-existing files would be
+	 * strange.
+	 */
+	/*om_notify_monitor (handle->uri, GNOME_VFS_MONITOR_EVENT_CHANGED);*/
+
+	return GNOME_VFS_OK;
 }
 
 static GnomeVFSResult 
@@ -1092,13 +1104,26 @@ do_get_file_info (GnomeVFSMethod          *method,
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 	
+	conn = om_get_connection (uri, &result);
+	if (!conn) {
+		return result;
+	}
+
 	/* Special-case the root. */
 	if (strcmp (path, "/") == 0) {
 		g_free (path);
-		
+
+		result = om_get_folder_listing (conn, uri, FALSE, context, 
+						NULL);
+		om_connection_unref (conn);
+
+		if (result != GNOME_VFS_OK) {
+			return result;
+		}
+
 		file_info->valid_fields = GNOME_VFS_FILE_INFO_FIELDS_NONE;
 		file_info->name = g_strdup ("/");
-		
+
 		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
 		file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
 
@@ -1110,11 +1135,6 @@ do_get_file_info (GnomeVFSMethod          *method,
 	
 	g_free (path);
 
-	conn = om_get_connection (uri, &result);
-	if (!conn) {
-		return result;
-	}
-
 	result = om_get_folder_listing (conn, uri, TRUE, context, &elements);
 	if (result != GNOME_VFS_OK) {
 		om_connection_unref (conn);
@@ -1193,6 +1213,9 @@ do_make_directory (GnomeVFSMethod  *method,
 	name = gnome_vfs_uri_extract_short_name (uri);
 	if (!gw_obex_mkdir (conn->obex, name, &error)) {
 		g_free (name);
+
+		om_connection_check_gwobex_error (conn, error);
+
 		om_set_cancel_context (conn, NULL);
 		om_connection_unref (conn);
 		return om_utils_obex_error_to_vfs_result (error);
@@ -1285,6 +1308,8 @@ do_unlink (GnomeVFSMethod  *method,
 	ret = gw_obex_delete (conn->obex, name, &error);
 	g_free (name);
 
+	om_connection_check_gwobex_error (conn, error);
+
 	om_connection_invalidate_cache (conn);
 
 	om_set_cancel_context (conn, NULL);
@@ -1393,6 +1418,8 @@ do_set_file_info (GnomeVFSMethod          *method,
 	ret = gw_obex_move (conn->obex, name, info->name, &error);
 	g_free (name);
 
+	om_connection_check_gwobex_error (conn, error);
+
 	om_set_cancel_context (conn, NULL);
 	om_connection_unref (conn);
 
@@ -1412,19 +1439,51 @@ do_set_file_info (GnomeVFSMethod          *method,
 	return GNOME_VFS_OK;
 }
 
-/* This is in gnome-vfs-monitor-private.h which is not installed, but it is
- * still part of the module API, so we need to have this hack here until that is
- * rectified in GnomeVFS.
- */
-void gnome_vfs_monitor_callback (GnomeVFSMethodHandle *method_handle,
-				 GnomeVFSURI *info_uri,
-				 GnomeVFSMonitorEventType event_type);
+static GnomeVFSResult
+do_get_volume_free_space (GnomeVFSMethod    *method,
+			  const GnomeVFSURI *uri,
+			  GnomeVFSFileSize  *free_space)
+{
+	GnomeVFSResult  result;
+	ObexConnection *conn;
+	gint            error;
+	gint            len;
+	gchar          *caps_str;
+	OvuCaps        *caps;
+	GError         *gerror = NULL;
+	GList          *entries;
 
+	conn = om_get_connection (uri, &result);
+	if (!conn) {
+		return result;
+	}
+
+	len = 0;
+	caps_str = NULL;
+	if (!gw_obex_get_capability (conn->obex, &caps_str, &len, &error)) {
+		om_connection_unref (conn);
+		return om_utils_obex_error_to_vfs_result (error);
+	}
+	
+	om_connection_unref (conn);
+
+	caps = ovu_caps_parser_parse (caps_str, len, &gerror);
+
+	entries = ovu_caps_get_memory_entries (caps);
+	if (entries) {
+		*free_space = ovu_caps_memory_get_free (entries->data);
+	}
+	
+	ovu_caps_free (caps);
+	
+	return GNOME_VFS_OK;
+}
 
 static void
 om_notify_monitor (GnomeVFSURI              *uri,
 		   GnomeVFSMonitorEventType  event_type)
 {
+	GnomeVFSURI   *tmp_uri;
 	MonitorList   *monitors;
 	GList         *l;
 	MonitorHandle *handle;
@@ -1436,12 +1495,19 @@ om_notify_monitor (GnomeVFSURI              *uri,
 	/* Note: gnome_vfs_monitor_callback can be called from any thread, so
 	 * there are no threading issues here.
 	 */
+	
+	tmp_uri = om_vfs_utils_create_uri_without_trailing_slash (uri);
+	monitors = g_hash_table_lookup (monitor_hash, tmp_uri);
+	gnome_vfs_uri_unref (tmp_uri);
 
-	monitors = g_hash_table_lookup (monitor_hash, uri);
 	if (monitors) {
 		for (l = monitors->handles; l; l = l->next) {
 			handle = l->data;
 
+			dv(g_printerr ("obex: emit %s event (direct) for %s\n",
+				       event_to_string (event_type),
+				       gnome_vfs_uri_to_string (uri, 0)));
+			
 			gnome_vfs_monitor_callback ((GnomeVFSMethodHandle *) handle,
 						    uri,
 						    event_type);
@@ -1453,13 +1519,20 @@ om_notify_monitor (GnomeVFSURI              *uri,
 		g_mutex_unlock (monitor_mutex);
 		return;
 	}
-	
-	monitors = g_hash_table_lookup (monitor_hash, parent);
+
+	tmp_uri = om_vfs_utils_create_uri_without_trailing_slash (parent);
+	monitors = g_hash_table_lookup (monitor_hash, tmp_uri);
+	gnome_vfs_uri_unref (tmp_uri);
+
 	if (monitors) {
 		for (l = monitors->handles; l; l = l->next) {
 			handle = l->data;
 
 			if (handle->monitor_type == GNOME_VFS_MONITOR_DIRECTORY) {
+				dv(g_printerr ("obex: emit %s event (parent) for %s\n",
+					       event_to_string (event_type),
+					       gnome_vfs_uri_to_string (uri, 0)));
+				
 				gnome_vfs_monitor_callback ((GnomeVFSMethodHandle *) handle,
 							    uri,
 							    event_type);
@@ -1480,13 +1553,19 @@ om_notify_monitor (GnomeVFSURI              *uri,
 		g_mutex_unlock (monitor_mutex);
 		return;
 	}
+
+	tmp_uri = om_vfs_utils_create_uri_without_trailing_slash (grand_parent);
+	monitors = g_hash_table_lookup (monitor_hash, tmp_uri);
+	gnome_vfs_uri_unref (tmp_uri);
 	
-	monitors = g_hash_table_lookup (monitor_hash, grand_parent);
 	if (monitors) {
 		for (l = monitors->handles; l; l = l->next) {
 			handle = l->data;
 
 			if (handle->monitor_type == GNOME_VFS_MONITOR_DIRECTORY) {
+				dv(g_printerr ("obex: emit changed event (grand parent) for %s\n",
+					       gnome_vfs_uri_to_string (parent, 0)));
+				
 				gnome_vfs_monitor_callback ((GnomeVFSMethodHandle *) handle,
 							    parent,
 							    GNOME_VFS_MONITOR_EVENT_CHANGED);
@@ -1510,7 +1589,7 @@ do_monitor_add (GnomeVFSMethod        *method,
 	MonitorList   *monitors;
 
 	handle = g_new0 (MonitorHandle, 1);
-	handle->uri = gnome_vfs_uri_dup (uri);
+	handle->uri = om_vfs_utils_create_uri_without_trailing_slash (uri);
 	handle->monitor_type = monitor_type;
 
 	g_mutex_lock (monitor_mutex);
@@ -1519,7 +1598,7 @@ do_monitor_add (GnomeVFSMethod        *method,
 	if (!monitors) {
 		monitors = g_new0 (MonitorList, 1);
 		g_hash_table_insert (monitor_hash,
-				     gnome_vfs_uri_dup (handle->uri),
+				     gnome_vfs_uri_ref (handle->uri),
 				     monitors);
 	}
 
@@ -1560,15 +1639,17 @@ do_monitor_cancel (GnomeVFSMethod       *method,
 	return GNOME_VFS_OK;
 }
 
-/* The vtable for vfs methods. The 5 unsupported methods are:
+/* The vtable for vfs methods. The unsupported methods are:
  *
- * truncate, truncate_handle:
+ * seek, tell, truncate, truncate_handle:
  * Those does not map well to the OBEX protocoll.
  *
  * find_directory:
  * This is used to search for a trash directory, and does not
  * make sense for a non-desktop filesystem.
  *
+ * forget_cache: doesn't apply here.
+ *
  * create_symbolic_link:
  * Not supported in the OBEX protocol.
 */
@@ -1580,8 +1661,8 @@ static GnomeVFSMethod method = {
         do_close,                     /* close */
         do_read,                      /* read */
         do_write,                     /* write */
-        do_seek,                      /* seek */
-        do_tell,                      /* tell */
+        NULL,                         /* seek */
+        NULL,                         /* tell */
         NULL,                         /* truncate_handle */
         do_open_directory,            /* open_directory */
         do_close_directory,           /* close_directory */
@@ -1600,7 +1681,9 @@ static GnomeVFSMethod method = {
         NULL,                         /* create_symbolic_link */
         do_monitor_add,               /* monitor_add */
         do_monitor_cancel,            /* monitor_cancel */
-        NULL                          /* file_control */
+        NULL,                         /* file_control */
+	NULL,                         /* forget_cache */
+	do_get_volume_free_space      /* get_volume_free_space */
 };
 
 static gboolean
@@ -1610,9 +1693,6 @@ get_caps_cb (const GnomeVFSURI *uri, gchar **caps, gint *len)
 	ObexConnection *conn;
 	gint            error;
 
-	/* For obex-less testing. */
-	/*return g_file_get_contents ("/home/rhult/CAP", caps, len, NULL);*/
-	
 	conn = om_get_connection (uri, &result);
 	if (conn == NULL) {
 		return FALSE;
@@ -1620,11 +1700,12 @@ get_caps_cb (const GnomeVFSURI *uri, gchar **caps, gint *len)
 			
 	*caps = NULL;
 	*len = 0;
-	
 	if (!gw_obex_get_capability (conn->obex,
 				     caps,
 				     len,
 				     &error)) {
+		om_connection_check_gwobex_error (conn, error);
+
 		om_connection_unref (conn);
 		return FALSE;
 	}
@@ -1649,7 +1730,7 @@ vfs_module_init (const char *method_name, const char *args)
 	if (!ovu_cap_server_init (get_caps_cb)) {
 		g_warning ("Couldn't init obex capabilities service.");
 	}
-	
+
 	return &method;
 }
 
